$ontext
This code writes script files to run a set of jobs on multiple
processors/machines plus GDX files to pass scenario data and scenario results
plus a GAMS file that runs the scenarios

It also writes a file merge_mergethem.gms that merges all into a unified file
and a script file that runs it

The concept is

This code - merge_control_example sets up the whole process.  It is told how to
allocate scenarios to processors/machines, is told what data to save and load.
Then it writes script files to be run on each processor/machine plus a gdx file
containing data, and a file that runs scenarios loads the scenario descriptions
then zips them to a file to be transferred to the processor/machine in turn on
the distributed processors the jobs are run and as they finish they unload the
results for the scenarios run into GDX files. Subsequently these GDX files need
to be transferred back to a location on which the data will be merged.

Operationally in the main merge_control_example.gms file we
    1.   first define the scenarios and in this example link them to some price,
         land and risk aversion modificiations (see section 1).

    2.   define the processors to be used and assign jobs that will be run by
         each processor (see section 2)

    3.   give a name to the procedure that will run the scenarios on the
         assigned processors (see section 3)

    4.   identify which parameters will be merged (see section 4 ).

    5.   set up the data items that will define the scenarios (see section 5)

    6.   Create processor specific GDX with scenario data to be run and scipt files
         that run it plus create a zip file to pass to the processor (Section 6).
         Subparts of this for each processor
         a.  set up data and put it in an appropriately named gdx file (section 6a)
         b.  create a script file for that processor (section 6b)
    7.   Create merge_mergethem.gms code that will merge the results (section 7)
         In creating that there are several steps that are labeled below
         a   define a control variable that contains the names of the report
             items to merge - section 7a
         b   declare the parameters and sets that are involved with holding the
             data to be merged (not needed if restarting) - section 7b
         c   include a loop to load gdx file data from each processor using
             $loadm so the data will be merged - section 7c
         d   form a set that tells what scenarios have been found - section 7e
         e   display merged results and write a gdx file of them  - section 7f
     8.  Get agreste.gms from the library and run it to create a restart file - section 8
     9.  Create GMS file that runs the scenarios called  agreste_loop - section 9
         a   set up scenario particulars - section 9a
         b   define the anmes of the cross scenario reports that willl be sent
             to be merged - section 9b
         c   load in data that tells what scenarios to use and accompanying data - section 9c
         d   for each scenario to run set it up, solve it and compute reports on
             the results - section 9d:
         e   unload the data to be merged into the gdx file - section 9e


B.  In doing this merge control_example writes several types of files
    1.   Files for use on remote processor
         a.  A script file for each processor with a name like
                  script_for_mainprocessor.gms.
             These are composed as gms files and need to be run on a remote processor.
             One will be generated for each processor identified in the runstodo table.
             It uses the 3 files below (b-d) after ther are unzipped from the
                  file discussed in e.
         b.  A GDX file for each processor with the scenario data named
                  send_to_processorname.gdx or in the example send_to_secondprocessor.gdx
         c.  A gms file that runs the scenarios called agreste_loop.gms.
         d.  A restart file that is created after agreste is run (a1.g00).
         e.  A zip file that contains the 4 files just above for each processor
                  called zip_to_send_to_processorname.zip or in the example
                  zip_to_send_to_Secondprocessor.zip.
    2.   A File for the merge operation named merge_mergethem.gms that loads the
         data in assuming the gdx files of results have been moved to the project
         directory.  It may be desirable to restart this file from after agreste
         or agreste_loop or merge_control_example providing access to problem data,
         sets etc. but for now we do not.

C.  On the remote processor one needs to
     1.  copy in and unzip the file zip_to_send_to_processorname.zip. In this
         example one is named zip_to_send_to_Secondprocessor.zip.  This unzips the
         script file (script_for_second_processor.gms, a1.goo, the agreste_loop.gms
         file and the send_to_processorname.gdx (e,g, send_to_secondprocessor.gdx)
     2.  running the script file causes it to load the gdx file of data,
         run agreste_loop.gms which executes all assigned scenarios and forms
         reports then saves the results in a gdx file for merger (results_processorname.gdx)
     3.  move the results gdx file to the location where the merge is to be run

D.  Then back on the main machine
    1.  one makes sure the results_processorname.gdx files have been moved
    2.  runs the merge_mergethem.gms to obtain the merged results. These
        will appear in merge_mergethem.lst or in results_agreste.gdx
    3.  Note this procedure can be run at any time even if all the gdx files are
        not present

$offtext

* Section 1: Define the scenarios to be run
set pricecases                   names of alternative price multipliers /pmult1*pmult5/;
parameter pricemult(pricecases)  data for alternative price multipliers/
         pmult1  0.8
         pmult2  0.9
         pmult3  1.0
         pmult4  1.1
         pmult5  1.2
/;

set landcases                    names of alternative land multipliers /Lmult1*Lmult3/;
parameter landmult(landcases)   data for alternative land multipliers/
         Lmult1  0.9
         Lmult2  1.0
         Lmult3  1.1
/;
set riskcases naems of alternative risk scenarios /r1,r2/;
parameter riskparams(riskcases) data for alternative risk parameters/
         r1  0
         r2  1
/;

* Define scenarios in terms of cases to be run
$eval maxscen card(pricecases)*card(landcases)*card(riskcases)
set scenarios scenarios to run /base,scen1*scen%maxscen%/;
set scenariossub(scenarios) /scen1*scen%maxscen%/
set linkscenarios(scenarios,pricecases,landcases,riskcases) tuple that links scenarios to cases
      / base.pmult3.lmult2.r1, #scenariossub:(#pricecases.#landcases.#riskcases) /;
option linkscenarios:0:0:4; display linkscenarios;

* Section 2: Define the processors that will be used and assign scenarios to processors
sets processor different processors / mainprocessor, Secondprocessor, thirdprocessor /;
* Tell which processor runs which scenarios
set runstodo(processor,scenarios)  scenario grouping for runs /
         mainprocessor.   (base,scen1*scen10)
         Secondprocessor. (scen11*scen20)
         thirdprocessor.  (scen21*scen30) /

* Note one can can use code like the following if one wishes to divide scenarios equally
$ontext
$eval jbatch ceil(card(scenarios)/card(processor))
sets runstodo(processor,scenarios)  processor scenario grouping for runs
     scenariosperprocessor scenarios per processor / b1*b%jbatch% /
     runstodoHelper(processor,scenariosperprocessor,scenarios) / (#processor.#scenariosperprocessor):#scenarios /
option runstodo<runstodoHelper;
$offtext

* Section 3: Give a name to the GAMS code that will run scenarios
$setglobal scenariosolver "agreste_loop"

* Section 4: Identify the names of the items that will be merged
$setglobal itemstomerge comparelanduse,runsummary,compareproduction

* Section 5: Set up the data items that will define the scenarios
set scen(scenarios)               names of the scenarios to be run by a processor;
parameter s_pricemult(scenarios)  values of price multipliers for the scenarios for a processor
          s_landmult(scenarios)   values of land multipliers for the scenarios for a processor,
          s_riskparams(scenarios) values of risk parameters for the scenarios for a processor;

* Section 6: Create GDX files with scenario data to be run for each processor
* Now put out data for a processor
file fx;
* Create GDX for each processor
loop(processor,

* Section 6a: Set up data and put it in an appropriately named GDX file

* Wipe out any old values
   option clear=s_pricemult, clear=s_landmult, clear=s_riskparams;

* Create items to pass with scenario names and data specific for the scenarios
*    to be run by this processor and data for those runs
   scen(scenarios)    = runstodo(processor, scenarios);
   s_pricemult(scen)  = sum(linkscenarios(scen,pricecases,landcases,riskcases), pricemult(pricecases));
   s_landmult(scen)   = sum(linkscenarios(scen,pricecases,landcases,riskcases), landmult(landcases));
   s_riskparams(scen) = sum(linkscenarios(scen,pricecases,landcases,riskcases), riskparams(riskcases));

* Set up the gdx file name for data to send to the processor
   put_utility fx 'gdxout' / 'send_to_' processor.tl:0 '.gdx';
* Put data into the gdx file
   execute_unload scen=scenarios, s_pricemult, s_landmult, s_riskparams;

* Section 6b: Create a script file for each processor

* Set up the name for the script which will be script_for_processorname.gms
   put_utility 'ren' / 'script_for_' processor.tl:0 '.gms'; fx.ap=0;
* Trick to stop evaluation of control variables
$escape =
* Now write contents of the script
   putclose '* script that Runs scenarios on a processor ' processor.tl:0
          / '$set processor ' processor.tl:0
          / '$if not exist a1.g00 $abort agreste restart file not present'
          / '$if not exist send_to_%=processor%=.gdx $abort data defining scenario for processor GDX file not present'
          / '$if not exist %scenariosolver%.gms $abort GAMS file that runs the scenario %scenariosolver%.gms file not present'
* Run the remote job starting from agreste save file and save results in restart file a2
          / '$call gams %scenariosolver% r=a1 --processor=%=processor%= --itemstomerge="%itemstomerge%" lo=3'
          / '$if errorlevel 1 $abort problems with running %scenariosolver%. Check %scenariosolver%.lst';
* Zip the results
   put_utility 'exec' / 'gmszip zip_to_send_to_' processor.tl:0 '.zip a1.g00 %scenariosolver%.gms script_for_' processor.tl:0 '.gms ' 'send_to_' processor.tl:0 '.gdx';
* Users might put a copy command here to move results to destination
* Users might put an unzip here to unzip files remotely
* Users could send a signal to begin the run
);

* Section 7: Create merge_mergethem.gms code that will merge the results
file fmergethem / 'merge_mergethem.gms' /; put fmergethem;

* Section 7a: Define a control variable that contains the names of the report items
*    to merge
put '$setglobal itemstomerge "%itemstomerge%"' /;

$onputS
* Enter base and scenario names in a junk set so it appears first as discussed
*    in set ordering part of McCarl guide
$if not declared putbasefirst set putbasefirst /base,scen1*scen%maxscen%/;
* Section 7b: Declare items involved with the parameters to summarize the results
*    that will be merged (not needed if restarting)
$onempty
$if not declared scenarios set scenarios(*) / /;
$offempty
$if not declared s alias (*,s);
$if not declared measures alias (*,measures);
$if not declared allcommod alias (*,allcommod);
         parameter comparelanduse(scenarios,s)             comparative land use report;
         parameter runsummary(scenarios,measures)          run summary stuff;
         parameter compareproduction(scenarios,allcommod)  compare production levels;
         set       processors_so_far(*)                    successful loads from processors;
         set       scenarios_so_far(*)                     successful loads of scenarios;

$offput
*Section 7c: Here is a loop to load gdx file data from each processor
loop(processor,
* First check if file is local
   put '$ifthen exist result_to_merge_' processor.tl:0 '.gdx '/;
   put '    $$gdxin "result_to_merge_' processor.tl:0 '.gdx"' /;
* Load it
   put '    $$loadM scenarios,%itemstomerge%' /;
   put '*   input file found for this processor so mark it as found' / ;
* Identify this one is done
   put '    processors_so_far("' processor.tl:0 '")=yes;' /;
   put '$endif' / /;
);

* Section 7e: Form a set that tells what scenarios have been found
$onputS

scenarios_so_far(scenarios)            = yes;

* Section 7f: Display merged results and write gdx file of them
* Display what we have merged so far
display processors_so_far,scenarios_so_far,%itemstomerge%;
* Put results in a gdx file
execute_unload "results_agreste.gdx", processors_so_far,scenarios_so_far,%itemstomerge%;

$offput

* Section 8: Load up agreste and create a restart file
$call gamslib agreste
$call gams agreste s=a1 lo=%gams.lo%

* Section 9: Create GAMS file for the agreste scenario run
$onechoV > agreste_loop.gms
* Section 9a: Set up scenario particulars
* Save data that will be changed
parameter saveprice(c)  save prices
          savelandc(s)  save land          ;
saveprice(c)=price(c);
savelandc(s)=landc(s);

* Scenario data
set scenarios set of the scenarios that have been run;
parameter s_pricemult(scenarios), s_landmult(scenarios), s_riskparams(scenarios);

* Section 9b: Define cross scenario reports that will be sent to be merged
alias (*,measures);
parameter comparelanduse(scenarios,s) comparative land use report;
parameter runsummary(scenarios,measures) run summary stuff;
set allcommod list of all commodities /set.c,livestock/ ;
parameter compareproduction(scenarios,allcommod)  compare production levels;

* Minimize output
option limrow=0,limcol=0,solprint=off;

* Section 9c: load in data that tells what scenarios to use and accompanying data
$gdxin send_to_%processor%.gdx
$load scenarios, s_pricemult, s_landmult, s_riskparams

* Section 9d: Run chosen scenarios
loop(scenarios,

* Update data to reflect scenario
    price(c) = saveprice(c)*s_pricemult(scenarios);
    landc(s) = savelandc(s)*s_landmult(scenarios);
    phi      = s_riskparams(scenarios);

* Solve
    solve agreste maximizing yfarm using lp;

* Build the tables of data to be merged
    comparelanduse(scenarios,s)=
           sum(p$ps(p,s), a(p)*xcrop.l(p,s))$sc(s)
         + sum(sp, ldp(s,sp)*lswitch.l(sp))
         + sum(r, lio(s,r)*xliver.l(r));
    runsummary(scenarios,"income")
         = revenue.l + vsc*sum(dr, cons.l(dr)) - labcost.l - rationr.l - vetcost.l - cropcost.l;
    runsummary(scenarios,"risk cost")
         = phi*sum(ty, pdev.l(ty) + ndev.l(ty))/card(ty);
    runsummary(scenarios,"mean abs dev")
         = sum(ty, pdev.l(ty) + ndev.l(ty))/card(ty);
    runsummary(scenarios,"Livestock prod")=xlive.l;
    compareproduction(scenarios,allcommod)=sum(sameas(allcommod,c),sum((s,p), yield(p,c,s)*xcrop.l(p,s)));
    compareproduction(scenarios,"livestock")=xlive.l;
    );

* Section 9e: Unload the data to be merged into the gdx file
* Build the file name as result_to_merge_processorname.gdx
execute_unload 'result_to_merge_%processor%.gdx', scenarios, %itemstomerge%
$offecho



